import funcnodes as fn
from funcnodes_images import ImageFormat
from typing import Optional, Tuple
from ._pillow import PillowImageFormat, NumpyImageFormat
from .utils import calc_crop_values
import numpy as np


class ShowImage(fn.Node):
    node_id = "image.show"
    node_name = "Show Image"

    default_render_options = {"data": {"src": "img", "type": "image"}}

    img = fn.NodeInput(
        id="img",
        type=ImageFormat,
    )

    async def func(self, img):
        pass


class ResizeImage(fn.Node):
    node_id = "image.resize"
    node_name = "Resize Image"

    img = fn.NodeInput(
        id="img",
        type=ImageFormat,
    )

    width = fn.NodeInput(
        id="width",
        type=int,
        required=False,
    )

    height = fn.NodeInput(
        id="height",
        type=int,
        required=False,
    )

    resized_img = fn.NodeOutput(
        id="resized_img",
        type=ImageFormat,
    )

    default_render_options = {"data": {"src": "resized_img", "type": "image"}}

    async def func(self, img: ImageFormat, width=None, height=None):
        out = img.resize(w=width, h=height)
        self.get_output("resized_img").value = out

        return out


class FromBytes(fn.Node):
    node_id = "image.from_bytes"
    node_name = "From Bytes"

    data = fn.NodeInput(
        id="data",
        type=bytes,
    )

    img = fn.NodeOutput(
        id="img",
        type=ImageFormat,
    )

    default_render_options = {"data": {"src": "img", "type": "image"}}

    async def func(self, data: bytes):
        img = PillowImageFormat.from_bytes(data)
        self.get_output("img").value = img
        return img


class ScaleImage(fn.Node):
    node_id = "image.scale"
    node_name = "Scale Image"

    img = fn.NodeInput(
        id="img",
        type=ImageFormat,
    )

    scale = fn.NodeInput(
        id="scale",
        type=float,
    )

    scaled_img = fn.NodeOutput(
        id="scaled_img",
        type=ImageFormat,
    )

    default_render_options = {"data": {"src": "scaled_img", "type": "image"}}

    async def func(self, img: ImageFormat, scale: float):
        out = img.scale(scale)
        self.get_output("scaled_img").value = out

        return out


class CropImage(fn.Node):
    node_id = "image.crop"
    node_name = "Crop Image"

    img = fn.NodeInput(
        id="img",
        type=ImageFormat,
    )

    x1 = fn.NodeInput(
        id="x1",
        type=int,
        name="left",
        default=0,
        required=False,
    )

    y1 = fn.NodeInput(
        id="y1",
        type=int,
        name="top",
        default=0,
        required=False,
    )

    x2 = fn.NodeInput(
        id="x2",
        type=int,
        name="right",
        required=False,
    )

    y2 = fn.NodeInput(
        id="y2",
        type=int,
        name="bottom",
        required=False,
    )

    cropped_img = fn.NodeOutput(
        id="cropped_img",
        type=ImageFormat,
    )

    default_render_options = {"data": {"src": "cropped_img", "type": "image"}}

    async def func(
        self,
        img: ImageFormat,
        x1: Optional[int] = None,
        y1: Optional[int] = None,
        x2: Optional[int] = None,
        y2: Optional[int] = None,
    ):
        x1, y1, x2, y2 = calc_crop_values(img.width(), img.height(), x1, y1, x2, y2)
        out = img.crop(x1, y1, x2, y2)

        self.get_input("x1").set_value(x1, does_trigger=False)
        self.get_input("y1").set_value(y1, does_trigger=False)
        self.get_input("x2").set_value(x2, does_trigger=False)
        self.get_input("y2").set_value(y2, does_trigger=False)
        self.get_output("cropped_img").value = out

        return out


class ToArray(fn.Node):
    node_id = "image.to_array"
    node_name = "To Array"

    img = fn.NodeInput(
        id="img",
        type=ImageFormat,
    )

    array = fn.NodeOutput(
        id="array",
        type="numpy.ndarray",
    )

    async def func(self, img: ImageFormat):
        out = img.to_array()
        self.get_output("array").value = out

        return out


class Dimensions(fn.Node):
    node_id = "image.dimensions"
    node_name = "Dimensions"

    img = fn.NodeInput(
        id="img",
        type=ImageFormat,
    )

    width = fn.NodeOutput(
        id="width",
        type=int,
    )

    height = fn.NodeOutput(
        id="height",
        type=int,
    )

    async def func(self, img: ImageFormat):
        self.get_output("width").value = img.width()
        self.get_output("height").value = img.height()
        return img.width(), img.height()


class FromArray(fn.Node):
    node_id = "image.from_array"
    node_name = "From Array"

    data = fn.NodeInput(
        id="data",
        type="numpy.ndarray",
    )

    img = fn.NodeOutput(
        id="img",
        type=ImageFormat,
    )

    default_render_options = {"data": {"src": "img", "type": "image"}}

    async def func(self, data):
        img = NumpyImageFormat(data)
        self.get_output("img").value = img
        return img


@fn.NodeDecorator(
    id="image.get_channels",
    name="Get Channels",
    outputs=[
        {
            "name": "red",
        },
        {
            "name": "green",
        },
        {
            "name": "blue",
        },
    ],
)
def get_channels(
    img: ImageFormat, as_rgb: bool = False
) -> Tuple[NumpyImageFormat, NumpyImageFormat, NumpyImageFormat]:
    rgb = img.to_np().data
    if rgb.shape[2] == 1:
        r = g = b = rgb[:, :, 0]
    else:
        r, g, b = rgb[:, :, 0], rgb[:, :, 1], rgb[:, :, 2]
    if not as_rgb:
        return NumpyImageFormat(r), NumpyImageFormat(g), NumpyImageFormat(b)
    rz = np.zeros_like(r, shape=(r.shape[0], r.shape[1], 3))
    rz[:, :, 0] = r
    gz = np.zeros_like(g, shape=(g.shape[0], g.shape[1], 3))
    gz[:, :, 1] = g
    bz = np.zeros_like(b, shape=(b.shape[0], b.shape[1], 3))
    bz[:, :, 2] = b

    return NumpyImageFormat(rz), NumpyImageFormat(gz), NumpyImageFormat(bz)


@fn.NodeDecorator(
    id="image.get_histograms",
    name="Get Histograms",
    outputs=[
        {
            "name": "red",
        },
        {
            "name": "green",
        },
        {
            "name": "blue",
        },
    ],
)
def histograms(img: ImageFormat) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    data = img.to_np().to_rgb_uint8()

    return (
        np.histogram(data[:, :, 0], bins=256, range=(0, 256))[0],
        np.histogram(data[:, :, 1], bins=256, range=(0, 256))[0],
        np.histogram(data[:, :, 2], bins=256, range=(0, 256))[0],
    )


@fn.NodeDecorator(
    id="image.to_jpeg",
    name="To JPEG",
    outputs=[
        {
            "name": "jpeg",
        },
    ],
    default_io_options={
        "quality": {"value_options": {"min": 0, "max": 100, "step": 1}}
    },
)
def to_jpeg(img: ImageFormat, quality: int) -> bytes:
    return img.to_jpeg(quality)


@fn.NodeDecorator(
    id="image.to_png",
    name="To PNG",
    outputs=[
        {
            "name": "png",
        },
    ],
)
def to_png(img: ImageFormat) -> bytes:
    return img.to_png()


NODE_SHELF = fn.Shelf(
    name="Images",
    nodes=[
        ShowImage,
        ResizeImage,
        FromBytes,
        ScaleImage,
        CropImage,
        ToArray,
        FromArray,
        Dimensions,
        get_channels,
        histograms,
        to_jpeg,
    ],
    subshelves=[],
    description="Basic Image processing nodes",
)



from PIL import Image
import numpy as np
from .imagecontainer import ImageFormat, register_imageformat
from ._numpy import NumpyImageFormat


class PillowImageFormat(ImageFormat[Image.Image]):
    def __init__(self, img: Image.Image) -> None:
        if not isinstance(img, Image.Image):
            raise TypeError("img must be a PIL Image")
        super().__init__(img)

    def to_array(self) -> np.ndarray:
        return np.array(self._data)

    def get_data_copy(self) -> Image.Image:
        return self._data.copy()

    def width(self) -> int:
        return self._data.width

    def height(self) -> int:
        return self._data.height


def pillow_to_numpy(img: PillowImageFormat) -> NumpyImageFormat:
    return NumpyImageFormat(np.array(img.data))


PillowImageFormat.add_to_converter(NumpyImageFormat, pillow_to_numpy)


def numpy_to_pil(img: NumpyImageFormat) -> PillowImageFormat:
    return PillowImageFormat(Image.fromarray(img.to_rgb_or_rgba_uint8()))


NumpyImageFormat.add_to_converter(PillowImageFormat, numpy_to_pil)


register_imageformat(PillowImageFormat, "img")



import unittest
from funcnodes_images.imagecontainer import (
    get_format,
)
from funcnodes_images._numpy import NumpyImageFormat
from funcnodes_images._pillow import PillowImageFormat
from PIL import Image
import numpy as np


class TestImageFormat(unittest.TestCase):
    def test_registration(self):
        f = get_format("np")
        self.assertEqual(f, NumpyImageFormat)
        f = get_format("img")
        self.assertEqual(f, PillowImageFormat)

    def test_to_array(self):
        # assuming conversion to NumpyImageFormat is correctly set up
        img = Image.new("RGB", (10, 10))
        pillow_format = PillowImageFormat(img)
        numpy_array = pillow_format.to_array()
        self.assertIsInstance(numpy_array, np.ndarray)

    def test_converter_registration_and_conversion(self):
        img = Image.new("RGB", (10, 10))
        pillow_format = PillowImageFormat(img)
        numpy_format = pillow_format.to("np")
        self.assertIsInstance(numpy_format, NumpyImageFormat)

    def test_to_thumbnail(self):
        img = Image.new("RGB", (100, 50))
        pillow_format = PillowImageFormat(img)
        thumbnail = pillow_format.to_thumbnail((50, 50))
        self.assertIsInstance(thumbnail.data, Image.Image)
        self.assertEqual(thumbnail.data.size, (50, 25))



from abc import ABC, abstractmethod
from typing import Type, Optional
import numpy as np
from typing import Any, Generic, TypeVar, Callable, Dict
import io
from PIL import Image
from .utils import calc_new_size, calc_crop_values

from funcnodes_core.config import update_render_options

T = TypeVar("T")


IMAGE_FORMATS: Dict[str, Type["ImageFormat"]] = {}


class ImageFormat(ABC, Generic[T]):  # noqa: F821
    _to_converters: Dict[
        "ImageFormat", Dict["ImageFormat", Callable[["ImageFormat"], "ImageFormat"]]
    ] = {}

    def __init__(self, data: T):
        self._data: T = data

    @property
    def data(self) -> T:
        return self.get_data_copy()

    @abstractmethod
    def get_data_copy(self) -> T:
        pass

    @abstractmethod
    def width(self) -> int:
        pass

    @abstractmethod
    def height(self) -> int:
        pass

    def to(self, cls: Type["ImageFormat"] | str) -> "ImageFormat":
        if isinstance(cls, str):
            cls = IMAGE_FORMATS[cls]
        if self.__class__ == cls:
            return self
        if cls in ImageFormat._to_converters[self.__class__]:
            return self._to_converters[self.__class__][cls](self)

    @classmethod
    def add_to_converter(
        self_cls,
        other_cls: Type["ImageFormat"],
        converter: Callable[["ImageFormat"], "ImageFormat"],
    ):
        if self_cls not in ImageFormat._to_converters:
            ImageFormat._to_converters[self_cls] = {}
        ImageFormat._to_converters[self_cls][other_cls] = converter

    def get_to_converter(
        self, cls: Type["ImageFormat"]
    ) -> Callable[["ImageFormat"], "ImageFormat"]:
        if self.__class__ == cls:
            return lambda x: x
        if self.__class__ not in ImageFormat._to_converters:
            ImageFormat._to_converters[self.__class__] = {}
        return ImageFormat._to_converters[self.__class__].get(cls)

    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name.startswith("to_"):
                _to = name[3:]

                if _to in IMAGE_FORMATS:
                    other = IMAGE_FORMATS[_to]
                    conv = self.get_to_converter(other)
                    if conv:
                        return lambda: conv(self)
            raise

    @classmethod
    def from_array(cls, data: np.ndarray):
        from ._numpy import NumpyImageFormat

        return NumpyImageFormat(data).to(cls)

    def to_array(self) -> np.ndarray:
        return self.to_np().data

    def __array__(self, dtype=None, copy=None):
        arr = self.to_array()
        if dtype:
            arr = arr.astype(dtype)
        if copy is not None and not copy:
            raise ValueError("copy=False is not supported")
        return arr

    @classmethod
    def from_file(cls, path: str):
        from ._pillow import PillowImageFormat

        img = Image.open(path)
        return PillowImageFormat(img).to(cls)

    @classmethod
    def from_bytes(cls, data: bytes):
        from ._pillow import PillowImageFormat

        buff = io.BytesIO(data)
        img = Image.open(buff)
        img.load()
        buff.close()
        return PillowImageFormat(img).to(cls)

    def to_jpeg(self, quality=75) -> bytes:
        img: Image = self.to_img().data
        img = img.convert("RGB")
        img_byte_array = io.BytesIO()
        img.save(img_byte_array, format="JPEG", quality=int(quality))
        return img_byte_array.getvalue()

    def to_png(self) -> bytes:
        img: Image = self.to_img().data
        img = img.convert("RGB")
        img_byte_array = io.BytesIO()
        img.save(img_byte_array, format="PNG")
        return img_byte_array.getvalue()

    def to_thumbnail(self, size: tuple) -> "ImageFormat[T]":
        img: Image = self.to_img().data
        img.thumbnail(size)
        return self.__class__.from_array(np.array(img))

    def resize(
        self,
        w: int = None,
        h: int = None,
    ) -> "ImageFormat[T]":
        img: Image = self.to_img().data
        new_x, new_y = calc_new_size(img.width, img.height, w, h)
        img = img.resize((new_x, new_y))
        return self.__class__.from_array(np.array(img))

    def crop(
        self,
        x1: Optional[int] = None,
        y1: Optional[int] = None,
        x2: Optional[int] = None,
        y2: Optional[int] = None,
    ) -> "ImageFormat[T]":
        img: Image = self.to_img().data
        x1, y1, x2, y2 = calc_crop_values(img.width, img.height, x1, y1, x2, y2)

        img = img.crop((x1, y1, x2, y2))
        return self.__class__.from_array(np.array(img))

    def scale(self, factor: float) -> "ImageFormat[T]":
        if factor <= 0:
            raise ValueError("factor must be greater than 0")

        img: Image = self.to_img().data
        img = img.resize((int(img.width * factor), int(img.height * factor)))
        return self.__class__.from_array(np.array(img))


def register_imageformat(
    imageformat: Type[ImageFormat], key: str, overwrite=False, _raise=True
):
    if not issubclass(imageformat, ImageFormat):
        raise ValueError("format must be a subclass of ImageFormat")
    if key in IMAGE_FORMATS:
        if IMAGE_FORMATS[key] == imageformat:
            return
        if not overwrite:
            if _raise:
                raise ValueError(
                    f"key '{key} 'already exists in image_formats as {IMAGE_FORMATS[key]}"
                )
            else:
                return
    IMAGE_FORMATS[key] = imageformat

    update_render_options(
        {
            "typemap": {
                imageformat: "image",
            }
        }
    )


def get_format(key: str) -> Type[ImageFormat]:
    return IMAGE_FORMATS[key]



from typing import Optional, Tuple


def calc_new_size(cur_x, cur_y, target_w: int, target_h: int) -> tuple:
    if target_w is not None and target_w < 1:
        target_w = None
    if target_h is not None and target_h < 1:
        target_h = None
    if target_w is None and target_h is None:
        raise ValueError("At least one of w or h must be given")
    if target_w is None:
        ratio = target_h / cur_y
        new_x, new_y = int(cur_x * ratio), target_h
    elif target_h is None:
        ratio = target_w / cur_x
        new_x, new_y = target_w, int(cur_y * ratio)
    else:
        new_x, new_y = target_w, target_h

    return new_x, new_y


def calc_crop_values(
    img_width: int,
    img_height: int,
    x1: Optional[int] = None,
    y1: Optional[int] = None,
    x2: Optional[int] = None,
    y2: Optional[int] = None,
) -> Tuple[int, int, int, int]:
    if x1 is None:
        x1 = 0
    if y1 is None:
        y1 = 0
    if x2 is None:
        x2 = img_width
    if y2 is None:
        y2 = img_height

    if x1 < 0 or x1 > img_width:
        x1 = x1 % img_width
    if y1 < 0 or y1 > img_height:
        y1 = y1 % img_height
    if x2 < 0 or x2 > img_width:
        x2 = x2 % img_width
    if y2 < 0 or y2 > img_height:
        y2 = y2 % img_height

    return x1, y1, x2, y2



from .imagecontainer import register_imageformat, get_format, ImageFormat
from ._numpy import NumpyImageFormat
from ._pillow import PillowImageFormat
import funcnodes as fn
from exposedfunctionality.function_parser.types import add_type
from . import image_nodes as nodes
import funcnodes_numpy as fn_numpy  # noqa: F401 # import for type hinting

add_type(ImageFormat, "ImageFormat")

FUNCNODES_RENDER_OPTIONS: fn.RenderOptions = {
    "typemap": {
        ImageFormat: "image",
    },
}


def imageFormatEncoder(obj: ImageFormat, preview=False):
    if isinstance(obj, ImageFormat):
        if preview:
            return obj.to_thumbnail((200, 200)).to_jpeg(), True
        return obj.to_jpeg(), True
    return obj, False


fn.JSONEncoder.add_encoder(imageFormatEncoder)


def imageFormatbytesEncoder(obj: ImageFormat, preview=False) -> fn.BytesEncdata:
    if isinstance(obj, ImageFormat):
        return fn.BytesEncdata(
            data=(obj.to_thumbnail((200, 200)) if preview else obj).to_jpeg(),
            mime="image/jpeg",
            handeled=True,
        )
    return fn.BytesEncdata(data=obj, handeled=False)


fn.ByteEncoder.add_encoder(imageFormatbytesEncoder, enc_cls=[ImageFormat])

NODE_SHELF = nodes.NODE_SHELF

__all__ = [
    "register_imageformat",
    "NumpyImageFormat",
    "get_format",
    "PillowImageFormat",
    "ImageFormat",
    "nodes",
    "FUNCNODES_RENDER_OPTIONS",
    "NODE_SHELF",
]


__version__ = "0.2.5"



import unittest
import funcnodes_images as fnimg
from funcnodes_images.imagecontainer import IMAGE_FORMATS
import numpy as np


class Test_IMAGE_FORMATS(unittest.TestCase):
    def test_image_formats(self):
        self.assertEqual(
            IMAGE_FORMATS,
            {"np": fnimg.NumpyImageFormat, "img": fnimg.PillowImageFormat},
        )


class TestNumpyImageFormat(unittest.TestCase):
    def test_format_create(self):
        data = np.random.rand(100, 100, 3)
        fmt = fnimg.get_format("np")(data)
        self.assertIsInstance(fmt, fnimg.NumpyImageFormat)

    def test_wrong_shape(self):
        data = np.random.rand(100, 100, 2)
        with self.assertRaises(ValueError):
            _ = fnimg.NumpyImageFormat(data)

    def test_to_uint8(self):
        data = np.random.rand(100, 100, 1)
        data[0, 0, 0] = 1
        data[1, 1, 0] = 0
        fmt = fnimg.NumpyImageFormat(data)

        self.assertIsInstance(fmt.to_uint8(), np.ndarray)
        self.assertEqual(fmt.to_uint8().max(), 255)
        self.assertEqual(fmt.to_uint8().min(), 0)

        data = np.random.rand(100, 100, 1)
        data[0, 0, 0] = 1
        data *= 100
        fmt = fnimg.NumpyImageFormat(data)
        self.assertIsInstance(fmt.to_uint8(), np.ndarray)
        self.assertEqual(fmt.to_uint8().max(), 100)
        self.assertEqual(fmt.to_uint8().min(), 0)

        data = np.random.rand(100, 100, 1)
        data[0, 0, 0] = 1
        data *= 300
        fmt = fnimg.NumpyImageFormat(data)
        self.assertIsInstance(fmt.to_uint8(), np.ndarray)
        self.assertEqual(fmt.to_uint8().max(), 255)
        self.assertEqual(fmt.to_uint8().min(), 0)

    def test_to_rgb_uint8(self):
        data = np.random.rand(100, 100, 1)
        fmt = fnimg.NumpyImageFormat(data)

        self.assertIsInstance(fmt.to_rgb_uint8(), np.ndarray)
        self.assertEqual(fmt.to_rgb_uint8().dtype, np.uint8)
        self.assertEqual(fmt.to_rgb_uint8().shape, (100, 100, 3))
        self.assertLessEqual(fmt.to_rgb_uint8().max(), 255)
        self.assertGreaterEqual(fmt.to_rgb_uint8().min(), 0)

    def test_short(self):
        data = np.random.rand(100, 100)
        fmt = fnimg.NumpyImageFormat(data)

        self.assertIsInstance(fmt, fnimg.NumpyImageFormat)
        self.assertEqual(fmt._data.shape, (100, 100, 1))



from all_nodes_test_base import TestAllNodesBase

import funcnodes_images as fnimg
import numpy as np
from PIL import Image
import tempfile
import io


class TestAllNodes(TestAllNodesBase):
    def setUp(self) -> None:
        self.img_arr = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
        self.img = Image.fromarray(self.img_arr)

    async def test_from_bytes(self):
        temp_jpg = tempfile.NamedTemporaryFile(suffix=".jpg")
        temp_jpg.close()
        self.img.save(temp_jpg.name, quality=95)
        with open(temp_jpg.name, "rb") as f:
            jpeg_bytes = f.read()

        refimg = Image.open(temp_jpg.name)

        frombytes = fnimg.nodes.FromBytes()
        frombytes.get_input("data").value = jpeg_bytes
        await frombytes
        img: fnimg.PillowImageFormat = frombytes.get_output("img").value
        self.assertIsInstance(img, fnimg.PillowImageFormat)

        np.testing.assert_equal(img.to_array(), np.array(refimg))

    async def test_resize(self):
        resize = fnimg.nodes.ResizeImage()
        resize.get_input("img").value = fnimg.PillowImageFormat(self.img)
        resize.get_input("width").value = 50
        resize.get_input("height").value = 50

        import asyncio

        await asyncio.sleep(1)
        img: fnimg.PillowImageFormat = resize.get_output("resized_img").value
        self.assertEqual(img.to_array().shape, (50, 50, 3))

    async def test_crop(self):
        crop = fnimg.nodes.CropImage()
        crop.get_input("img").value = fnimg.PillowImageFormat(self.img)
        crop.get_input("x1").value = 10
        crop.get_input("y1").value = 10
        crop.get_input("x2").value = 90
        crop.get_input("y2").value = 90

        await crop
        img: fnimg.PillowImageFormat = crop.get_output("cropped_img").value
        self.assertEqual(img.to_array().shape, (80, 80, 3))

    async def test_scale(self):
        scale = fnimg.nodes.ScaleImage()
        scale.get_input("img").value = fnimg.PillowImageFormat(self.img)
        scale.get_input("scale").value = 0.5

        await scale
        img: fnimg.PillowImageFormat = scale.get_output("scaled_img").value
        self.assertEqual(img.to_array().shape, (50, 50, 3))

    async def test_get_channels_from_rgb(self):
        get_channels = fnimg.nodes.get_channels()
        # testnormal
        get_channels.get_input("img").value = fnimg.PillowImageFormat(self.img)
        await get_channels
        red: fnimg.NumpyImageFormat = get_channels.get_output("red").value
        green: fnimg.NumpyImageFormat = get_channels.get_output("green").value
        blue: fnimg.NumpyImageFormat = get_channels.get_output("blue").value

        self.assertEqual(red.to_array().shape, (100, 100, 1))
        self.assertEqual(green.to_array().shape, (100, 100, 1))
        self.assertEqual(blue.to_array().shape, (100, 100, 1))
        np.testing.assert_equal(red.to_array(), self.img_arr[:, :, [0]])
        np.testing.assert_equal(green.to_array(), self.img_arr[:, :, [1]])
        np.testing.assert_equal(blue.to_array(), self.img_arr[:, :, [2]])

        get_channels.get_input("img").value = fnimg.PillowImageFormat(self.img)
        get_channels.get_input("as_rgb").value = True
        await get_channels
        red: fnimg.NumpyImageFormat = get_channels.get_output("red").value
        green: fnimg.NumpyImageFormat = get_channels.get_output("green").value
        blue: fnimg.NumpyImageFormat = get_channels.get_output("blue").value

        self.assertEqual(red.to_array().shape, (100, 100, 3))
        self.assertEqual(green.to_array().shape, (100, 100, 3))
        self.assertEqual(blue.to_array().shape, (100, 100, 3))

        self.assertTrue(np.all(red.to_array()[:, :, 0] == self.img_arr[:, :, 0]))
        self.assertTrue(np.all(green.to_array()[:, :, 1] == self.img_arr[:, :, 1]))
        self.assertTrue(np.all(blue.to_array()[:, :, 2] == self.img_arr[:, :, 2]))
        self.assertTrue(np.all(red.to_array()[:, :, 1] == 0))
        self.assertTrue(np.all(red.to_array()[:, :, 2] == 0))
        self.assertTrue(np.all(green.to_array()[:, :, 0] == 0))
        self.assertTrue(np.all(green.to_array()[:, :, 2] == 0))
        self.assertTrue(np.all(blue.to_array()[:, :, 0] == 0))
        self.assertTrue(np.all(blue.to_array()[:, :, 1] == 0))

    async def test_get_channels_from_bw(self):
        # bw image
        get_channels = fnimg.nodes.get_channels()
        get_channels.get_input("img").value = fnimg.NumpyImageFormat(
            self.img_arr[:, :, 0]
        )
        await get_channels
        red: fnimg.NumpyImageFormat = get_channels.get_output("red").value
        green: fnimg.NumpyImageFormat = get_channels.get_output("green").value
        blue: fnimg.NumpyImageFormat = get_channels.get_output("blue").value

        self.assertEqual(red.to_array().shape, (100, 100, 1))
        self.assertEqual(green.to_array().shape, (100, 100, 1))
        self.assertEqual(blue.to_array().shape, (100, 100, 1))
        np.testing.assert_equal(red.to_array(), self.img_arr[:, :, [0]])
        np.testing.assert_equal(green.to_array(), self.img_arr[:, :, [0]])
        np.testing.assert_equal(blue.to_array(), self.img_arr[:, :, [0]])

        get_channels.get_input("img").value = fnimg.NumpyImageFormat(
            self.img_arr[:, :, 0]
        )
        get_channels.get_input("as_rgb").value = True
        await get_channels
        red: fnimg.NumpyImageFormat = get_channels.get_output("red").value
        green: fnimg.NumpyImageFormat = get_channels.get_output("green").value
        blue: fnimg.NumpyImageFormat = get_channels.get_output("blue").value

        self.assertEqual(red.to_array().shape, (100, 100, 3))
        self.assertEqual(green.to_array().shape, (100, 100, 3))
        self.assertEqual(blue.to_array().shape, (100, 100, 3))

        np.testing.assert_equal(green.to_array()[:, :, 1], red.to_array()[:, :, 0])
        np.testing.assert_equal(blue.to_array()[:, :, 2], red.to_array()[:, :, 0])
        np.testing.assert_equal(green.to_array()[:, :, 1], blue.to_array()[:, :, 2])

        self.assertTrue(np.all(red.to_array()[:, :, 0] == self.img_arr[:, :, 0]))

        self.assertTrue(np.all(green.to_array()[:, :, 1] == self.img_arr[:, :, 0]))
        self.assertTrue(np.all(blue.to_array()[:, :, 2] == self.img_arr[:, :, 0]))
        self.assertTrue(np.all(red.to_array()[:, :, 1] == 0))
        self.assertTrue(np.all(red.to_array()[:, :, 2] == 0))
        self.assertTrue(np.all(green.to_array()[:, :, 0] == 0))
        self.assertTrue(np.all(green.to_array()[:, :, 2] == 0))
        self.assertTrue(np.all(blue.to_array()[:, :, 0] == 0))
        self.assertTrue(np.all(blue.to_array()[:, :, 1] == 0))

    async def test_get_histograms(self):
        hist = fnimg.nodes.histograms()
        hist.get_input("img").value = fnimg.PillowImageFormat(self.img)
        await hist
        red: np.ndarray = hist.get_output("red").value
        green: np.ndarray = hist.get_output("green").value
        blue: np.ndarray = hist.get_output("blue").value

        self.assertEqual(red.shape, (256,))
        self.assertEqual(green.shape, (256,))
        self.assertEqual(blue.shape, (256,))

        self.assertEqual(red.sum(), 10000)
        self.assertEqual(green.sum(), 10000)
        self.assertEqual(blue.sum(), 10000)

        np.testing.assert_equal(
            red, np.histogram(self.img_arr[:, :, 0], bins=256, range=(0, 255))[0]
        )
        np.testing.assert_equal(
            green, np.histogram(self.img_arr[:, :, 1], bins=256, range=(0, 255))[0]
        )
        np.testing.assert_equal(
            blue, np.histogram(self.img_arr[:, :, 2], bins=256, range=(0, 255))[0]
        )

    async def test_from_array(self):
        fromarray = fnimg.nodes.FromArray()
        fromarray.get_input("data").value = self.img_arr
        await fromarray
        img: fnimg.NumpyImageFormat = fromarray.get_output("img").value
        self.assertEqual(img.to_array().shape, self.img_arr.shape)
        np.testing.assert_equal(img.to_array(), self.img_arr)

    async def test_dimensions(self):
        dim = fnimg.nodes.Dimensions()
        dim.get_input("img").value = fnimg.PillowImageFormat(self.img)
        await dim
        width: int = dim.get_output("width").value
        height: int = dim.get_output("height").value

        self.assertEqual(width, 100)
        self.assertEqual(height, 100)

    async def test_to_array(self):
        toarray = fnimg.nodes.ToArray()
        toarray.get_input("img").value = fnimg.PillowImageFormat(self.img)
        await toarray
        arr: np.ndarray = toarray.get_output("array").value

        self.assertEqual(arr.shape, self.img_arr.shape)
        np.testing.assert_equal(arr, self.img_arr)
        np.testing.assert_equal(np.asarray(self.img), self.img_arr)

    async def test_show_image(self):
        show = fnimg.nodes.ShowImage()
        show.get_input("img").value = fnimg.PillowImageFormat(self.img)
        await show

    async def test_to_jpeg(self):
        tojpeg = fnimg.nodes.to_jpeg()
        tojpeg.get_input("img").value = fnimg.PillowImageFormat(self.img)
        tojpeg.get_input("quality").value = 95
        await tojpeg
        # jpeg: bytes = tojpeg.get_output("jpeg").value

    async def test_to_png(self):
        topng = fnimg.nodes.to_png()
        topng.get_input("img").value = fnimg.PillowImageFormat(self.img)
        await topng
        with Image.open(io.BytesIO(topng["png"].value)) as img:
            self.assertEqual(img.format, "PNG")



# this is an autogenerated file, do not edit it directly or your changes might be lost.
import unittest
from funcnodes_images import NODE_SHELF  # noqa
from functools import wraps
from typing import List
import funcnodes as fn
import asyncio


def passfunc(self, *args, **kwargs):
    pass


async def async_passfunc(self, *args, **kwargs):
    pass


def add_subclass_tests(cls):
    # Dynamically add test methods from sub_test_classes
    if not hasattr(cls, "sub_test_classes"):
        return
    for testcase in cls.sub_test_classes:
        if hasattr(testcase, "setUp"):
            inner_setup = testcase.setUp
        else:
            inner_setup = passfunc

        if hasattr(testcase, "asyncSetUp"):
            inner_async_setup = testcase.asyncSetUp
        else:
            inner_async_setup = async_passfunc

        if hasattr(testcase, "tearDown"):
            inner_teardown = testcase.tearDown
        else:
            inner_teardown = passfunc

        if hasattr(testcase, "asyncTearDown"):
            inner_async_teardown = testcase.asyncTearDown
        else:
            inner_async_teardown = async_passfunc

        for attr_name in dir(testcase):
            if attr_name.startswith("test_"):
                # Retrieve the test method from the subclass
                test_method = getattr(testcase, attr_name)

                # Create a new test method that wraps the original test method
                def make_new_test_method(
                    test_method=test_method,
                    inner_setup=inner_setup,
                    inner_teardown=inner_teardown,
                    inner_async_setup=inner_async_setup,
                    inner_async_teardown=inner_async_teardown,
                ):
                    if asyncio.iscoroutinefunction(test_method):

                        async def test_method_wrapper(self, *args, **kwargs):
                            # Call the inner setup method
                            inner_setup(self)
                            await inner_async_setup(self)
                            # Call the test method
                            await test_method(self, *args, **kwargs)
                            print(f"Test {test_method.__name__} passed")
                            # Call the inner teardown method
                            inner_teardown(self)
                            await inner_async_teardown(self)

                    else:

                        def test_method_wrapper(self, *args, **kwargs):
                            # Call the inner setup method
                            inner_setup(self)
                            # Call the test method
                            test_method(self, *args, **kwargs)
                            print(f"Test {test_method.__name__} passed")
                            # Call the inner teardown method
                            inner_teardown(self)

                    return test_method_wrapper

                # Create a unique name for the test method in this class
                new_test_name = f"test_{testcase.__name__}_{attr_name}"
                # Add the test method to this class
                setattr(
                    cls,
                    new_test_name,
                    make_new_test_method(
                        test_method=test_method,
                        inner_setup=inner_setup,
                        inner_teardown=inner_teardown,
                        inner_async_setup=inner_async_setup,
                        inner_async_teardown=inner_async_teardown,
                    ),
                )


class TestAllNodesBase(unittest.IsolatedAsyncioTestCase):
    # if you tests your nodes with in other test classes, add them here
    # this will automtically extend this test class with the tests in the other test classes
    # but this will also mean if you run all tests these tests might run multiple times
    # also the correspondinig setups and teardowns will not be called, so the tests should be
    # independently callable
    sub_test_classes: List[unittest.IsolatedAsyncioTestCase] = []

    # if you have specific nodes you dont want to test, add them here
    # But why would you do that, it will ruin the coverage?!
    # a specific use case would be ignore nodes that e.g. load a lot of data, but there we would recommend
    # to write tests with patches and not ignore them.
    ignore_nodes: List[fn.Node] = []

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        add_subclass_tests(cls)
        return cls

    @classmethod
    def setUpClass(cls):
        from funcnodes_core import testing

        testing.setup()

        def get_all_nodes_classes(shelf: fn.Shelf, current=None):
            if current is None:
                current = []
            for node in shelf.nodes:
                if node not in current:
                    current.append(node)

            for subshelf in shelf.subshelves:
                get_all_nodes_classes(subshelf, current)

            return current

        all_nodes = fn.flatten_shelf(NODE_SHELF)[0]
        nodes_to_test = all_nodes.copy()

        for node in cls.ignore_nodes:
            if node in nodes_to_test:
                nodes_to_test.remove(node)

        # monkey patching the async func method in the nodeclasses that if they are called the nodeclass
        # is removed from the NODES_TO_TEST list
        def monkey_patch_func(node_class):
            ofunc = node_class.func
            node_class.TestAllNodes_func = ofunc

            @wraps(ofunc)
            async def func(self, *args, **kwargs):
                res = await ofunc(self, *args, **kwargs)
                if node_class in nodes_to_test:
                    nodes_to_test.remove(node_class)
                return res

            node_class.func = func

        for node_class in all_nodes:
            monkey_patch_func(node_class)

        cls.all_nodes = all_nodes
        cls.nodes_to_test = nodes_to_test

    @classmethod
    def tearDownClass(cls):
        # undo the monkey patching
        from funcnodes_core import testing

        testing.teardown()

        for node_class in cls.all_nodes:
            if hasattr(node_class, "TestAllNodes_func"):
                node_class.func = node_class.TestAllNodes_func
                del node_class.TestAllNodes_func

        # Final assertion to ensure all nodes were tested
        if cls.nodes_to_test:
            raise AssertionError(
                f"These nodes were not tested ({len(cls.nodes_to_test)}): {cls.nodes_to_test}"
            )



import unittest
from funcnodes_images._numpy import NumpyImageFormat
import numpy as np


class TestNumpyImageFormat(unittest.TestCase):
    def test_initialization(self):
        arr = np.zeros((10, 10, 3))
        img_format = NumpyImageFormat(arr)
        self.assertIsInstance(img_format, NumpyImageFormat)

    def test_bad_initialization(self):
        with self.assertRaises(ValueError):
            NumpyImageFormat(np.zeros((10, 10, 2)))  # 2 channels

    def test_to_uint8(self):
        arr = np.random.rand(10, 10, 3) * 300  # values greater than 255
        img_format = NumpyImageFormat(arr)
        uint8_array = img_format.to_uint8()
        self.assertTrue(np.all(uint8_array <= 255))
        self.assertTrue(np.all(uint8_array >= 0))
        self.assertEqual(uint8_array.dtype, np.uint8)


# Additional tests for dimension expansion, channel checks, etc.



from .imagecontainer import ImageFormat, register_imageformat


import numpy as np


class NumpyImageFormat(ImageFormat[np.ndarray]):
    def __init__(self, arr: np.ndarray) -> None:
        # check if arr is a numpy array
        if not isinstance(arr, np.ndarray):
            raise TypeError("arr must be a numpy array")
        # make shure arr has 3 dimensions or fail

        if len(arr.shape) != 3:
            if len(arr.shape) == 2:
                arr = np.expand_dims(arr, axis=2)
            else:
                raise ValueError("arr must have 3 dimensions")

        # allow 3 or 1 channel images
        if arr.shape[2] != 3 and arr.shape[2] != 1 and arr.shape[2] != 4:
            raise ValueError(
                f"arr must have 1,3 or 4 channels but has shape {arr.shape}"
            )

        super().__init__(arr)

    def get_data_copy(self) -> np.ndarray:
        return self._data.copy()

    def width(self) -> int:
        return self._data.shape[1]

    def height(self) -> int:
        return self._data.shape[0]

    def to_uint8(self) -> np.ndarray:
        d = self.data
        if d.dtype == np.uint8:
            return d

        _max = d.max()
        _min = d.min()

        # if in range 0-255
        if _max <= 255 and _min >= 0 and _max > 1:
            return d.astype(np.uint8)

        # if not in range 0-1 -> normalize
        if _max > 1 or _min < 0:
            d = d - _min
            if _max != _min:
                d = d / (_max - _min)

        return (d * 255).astype(np.uint8)

    def to_rgb_or_rgba_uint8(self) -> np.ndarray:
        d = self.to_uint8()
        if d.shape[2] == 3 or d.shape[2] == 4:
            return d

        return np.repeat(d, 3, axis=2)

    def to_rgb_uint8(self) -> np.ndarray:
        d = self.to_uint8()
        if d.shape[2] == 3:
            return d

        if d.shape[2] == 4:
            return d[:, :, :3]

        return np.repeat(d, 3, axis=2)

    def to_rgba_uint8(self) -> np.ndarray:
        d = self.to_uint8()
        if d.shape[2] == 4:
            return d

        if d.shape[2] == 3:
            return np.concatenate(
                [d, np.full(d.shape[:2] + (1,), 255, dtype=np.uint8)], axis=2
            )

        return np.repeat(d, 4, axis=2)

    def to_array(self) -> np.ndarray:
        return self.get_data_copy()


register_imageformat(NumpyImageFormat, "np")



import unittest
from funcnodes_images._pillow import PillowImageFormat
from PIL import Image
import numpy as np


class TestPillowImageFormat(unittest.TestCase):
    def test_initialization(self):
        img = Image.new("RGB", (10, 10))
        img_format = PillowImageFormat(img)
        self.assertIsInstance(img_format, PillowImageFormat)

    def test_to_array(self):
        img = Image.new("RGB", (10, 10))
        img_format = PillowImageFormat(img)
        arr = img_format.to_array()
        self.assertIsInstance(arr, np.ndarray)
        self.assertEqual(arr.shape, (10, 10, 3))


# Additional tests for conversions, error handling on bad input, etc.
